/*
 * package: SAMLPostAdaptor
 *
 * $Id: SAMLPostAdaptor.java,v 1.3 2005/05/26 12:22:22 macias Exp $
 */
package es.rediris.aarr.protocol.saml;

import es.rediris.aarr.Diagnostic;
import es.rediris.aarr.ProfileManager;
import es.rediris.aarr.data.ObservableVector;
import es.rediris.aarr.protocol.saml.DataUnit;
import es.rediris.aarr.protocol.ProtocolAdaptor;
import es.rediris.aarr.protocol.ProtocolAdaptorConfigException;
import es.rediris.aarr.protocol.ProtocolAdaptorReadException;
import es.rediris.aarr.protocol.ProtocolAdaptorSendException;

import org.apache.xml.security.c14n.CanonicalizationException;
import org.apache.xml.security.c14n.Canonicalizer;
import org.apache.xml.security.c14n.InvalidCanonicalizerException;

import org.opensaml.SAMLAction;
import org.opensaml.SAMLAssertion;
import org.opensaml.SAMLAttribute;
import org.opensaml.SAMLAttributeDesignator;
import org.opensaml.SAMLAttributeQuery;
import org.opensaml.SAMLAttributeStatement;
import org.opensaml.SAMLAuthenticationQuery;
import org.opensaml.SAMLAuthenticationStatement;
import org.opensaml.SAMLAuthorizationDecisionQuery;
import org.opensaml.SAMLAuthorizationDecisionStatement;
import org.opensaml.SAMLDecision;
import org.opensaml.SAMLException;
import org.opensaml.SAMLQuery;
import org.opensaml.SAMLRequest;
import org.opensaml.SAMLResponse;
import org.opensaml.SAMLSOAPBinding;
import org.opensaml.SAMLStatement;
import org.opensaml.SAMLSubject;
import org.opensaml.SOAPException;
import org.opensaml.XML;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import org.xml.sax.SAXException;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import java.net.ServerSocket;
import java.net.Socket;

import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.Observer;
import java.util.StringTokenizer;
import java.util.Vector;


/**
 * This class extends <code>org.opensaml.SAMLSOAPBinding</code> and implements interfaces
 * {@link ProtocolAdaptor ProtocolAdaptor},  <code>Runnable</code> and <code>Observer</code>.
 *
 * @author kan
 */
public class SAMLPostAdaptor extends SAMLSOAPBinding implements ProtocolAdaptor,
    Runnable {
    /** The SAML protocol */
    public final static String SAML_PROTOCOL = "saml";

    /** A thread */
    Thread thread;

    /** An element that will store a configured protocol */
    Element configProtocol;

    /** A server socket*/
    ServerSocket serverSocket;

    /** An {@link ObservableVector ObservableVector} used to store queries */
    ObservableVector inputQueue;

    /** A {@link ProfileManager ProfileManager} object */
    ProfileManager profileManager;

    /** The issuer's name */
    String issuerName;

    /** This method is defined in interface {@link ProtocolAdaptor ProtocolAdaptor} and is used
     * to perform the following tasks:
     * <ul>
     * <li>Sets the profileManager for the SAMLPostAdaptor with the given parameter
     * <li>Try to:
     * <ul>
     * <li>Create a new socket listening at port 80
     * <li>Start a new thread
     * </ul>
     * <li>Catches an <code>IOException</code> exception if it's not able to create the server socket or start a new thread
     * </ul>
     *
     * @param profileManager A {@link ProfileManager ProfileManager}*/
    public void init(ProfileManager profileManager) {
        inputQueue = new ObservableVector();

        try {
            serverSocket = new ServerSocket(80);
            thread = new Thread(this, "ProtocolAdaptor");
            thread.start();
        } catch (IOException e) {
            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                "IOException: " + e.getMessage());
        }
    }

    /** This method wait for connections from a client, {@link SAMLPostAdaptor#receive receives} a SAML request, and
     * creates a {@link DataUnit DataUnit} with the accepted client socket and the
     * SAML request.  */
    public void run() {
        while (true) {
            Socket clientSocket = null;

            try {
                int contentLength = 0;
                clientSocket = serverSocket.accept();
                Diagnostic.writeLog(Diagnostic.INFO_MODE,
                    "Client connection from " +
                    clientSocket.getInetAddress().toString());

                BufferedReader in = new BufferedReader(new InputStreamReader(
                            clientSocket.getInputStream()));
                String inputLine = "";
                boolean finalHeader = false;

                while (!finalHeader && ((inputLine = in.readLine()) != null)) {
                    if (inputLine.equals("")) {
                        finalHeader = true;
                    } else if (inputLine.startsWith("Content-Length:")) {
                        StringTokenizer st = new StringTokenizer(inputLine);
                        st.nextToken();
                        contentLength = Integer.parseInt(st.nextToken());
                    }
                }

                String msg = "HTTP/1.0 200 OK";
                byte[] bytes = new byte[contentLength];
                clientSocket.getInputStream().read(bytes);

                String message = new String(bytes);

                SAMLRequest request = receive(message);

                DataUnit currentRequest = new DataUnit(clientSocket, request);

                // Add elements to observable vector query
                inputQueue.addElement(currentRequest);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    /**  This method parses a SOAP message and returns a SAMLRequest object on success. Otherwise
     * returns an exception related to one of these possible errors:
     * <ul>
     * <li>incompatible or missing SOAP envelope
     * <li>a mandatory SOAP header is [not?] detected
     * <li>XML parsing error
     * <li>I/O Error
     * <li>Unknown error
     * </ul>
     * @param soapMessage a SOAP message
     * @return a {@link SAMLRequest SAMLRequest} object
     * @throws a <code>SAMLException</code> on errors. These errors can be:  */
    public SAMLRequest receive(Object soapMessage) throws SAMLException {
        try {
            InputStream byteData = new ByteArrayInputStream(((String) soapMessage).getBytes());
            Document doc = XML.parserPool.parse(byteData);

            Element e = doc.getDocumentElement();

            // The root must be a SOAP 1.1 envelope.
            if (!XML.isElementNamed(e, XML.SOAP11ENV_NS, "Envelope")) {
                throw new SOAPException(SOAPException.VERSION,
                    "SAMLPostAdaptor.receive() detected an incompatible or missing SOAP envelope");
            }

            Node n = e.getFirstChild();

            while ((n != null) && (n.getNodeType() != Node.ELEMENT_NODE))
                n = n.getNextSibling();

            if (XML.isElementNamed((Element) n, XML.SOAP11ENV_NS, "Header")) {
                Node header = n.getFirstChild();

                while (header != null) {
                    if ((header.getNodeType() == Node.ELEMENT_NODE) &&
                            ((Element) header).getAttributeNS(
                                XML.SOAP11ENV_NS, "mustUnderstand").equals("1")) {
                        throw new SOAPException(SOAPException.MUSTUNDERSTAND,
                            "SAMLPostAdaptor.receive() detected a mandatory SOAP header");
                    }

                    header = header.getNextSibling();
                }

                // Advance to the Body element.
                n = n.getNextSibling();

                while ((n != null) && (n.getNodeType() != Node.ELEMENT_NODE))
                    n = n.getNextSibling();
            }

            if (n != null) {
                n = n.getFirstChild();

                while ((n != null) && (n.getNodeType() != Node.ELEMENT_NODE))
                    n = n.getNextSibling();
            }

            return new SAMLRequest((Element) n);
        } catch (SAXException e) {
            throw new SOAPException(SOAPException.CLIENT,
                "SAMLPostAdaptor.receive() detected an XML parsing error: " +
                e.getMessage());
        } catch (java.io.IOException e) {
            throw new SOAPException(SOAPException.SERVER,
                "SAMLPostAdaptor.receive() detected an I/O error: " +
                e.getMessage());
        } catch (Exception e) {
            throw new SOAPException(SOAPException.SERVER,
                "SAMLPostAdaptor.receive() detected an error: " +
                e.getMessage());
        }
    }

    /** This method returns SAML responses (<code>SAMLResponse</code> objects) depending on the <code>query</code> param passed and the
     * data provided . These responses can be of three different types:
     * <ul>
     * <li>SAML attributes
     * <li>SAML authorization decisions
     * <li>SAML authentication decisions
     * </ul>
     *
     * Methods {@link SAMLPostAdaptor#createAuthorizationResponse(DataUnit query, String [] ids, String [] values) createAuthorizationResponse},
     * {@link SAMLPostAdaptor#createAttributeResponse( DataUnit query, String [] ids, String [] values) createAttributeResponse} and
     * {@link  SAMLPostAdaptor#createAuthenticationResponse( DataUnit query, String [ ] ids, String [] values) createAuthenticationResponse} are
     * invoked depending on the <code>query</code> param.
     *
     * @param query A {@link DataUnit DataUnit} object
     * @param ids An array of ids needed for creating a response
     * @param values An array of ids needed linked with the ids, and also needed for creating the response
     *
     * @return a {@link SAMLResponse SAMLResponse} object
     * @throws UnknownUnit when the query is not applicable to the SAML protocol
     *
     * @see es.rediris.aarr.protocol.saml#sendElement(String name, String[] components, String[] values) */
    private SAMLResponse createResponse(DataUnit query, String[] ids,
        String[] values) throws InvalidSAMLRequestException {
        DataUnit theQuery = (DataUnit) query;
        SAMLRequest sreq = (SAMLRequest) theQuery.getPacket();
        SAMLQuery q = sreq.getQuery();

        if (q instanceof SAMLAttributeQuery) {
            return createAttributeResponse(query, ids, values);
        } else if (q instanceof SAMLAuthorizationDecisionQuery) {
            return createAuthorizationResponse(query, ids, values);
        } else if (q instanceof SAMLAuthenticationQuery) {
            return createAuthenticationResponse(query, ids, values);
        }

        throw new InvalidSAMLRequestException("This request (" + q.getClass().getName() +
            ") is not aplicable to SAML protocol");
    }

    /** Returns a <code>SAMLResponse(String inResponseTo, String recipient, Collection assertions, SAMLException e)</code> in response
     * to an authorization query, by taking into account the provided data.
     *
     * @param query A {@link DataUnit DataUnit} object
     * @param ids An array of ids needed for creating a response
     * @param values An array of ids needed linked with the ids, and also needed for creating the response
     *
     * @return a <code>SAMLResponse</code> object
     */
    private SAMLResponse createAuthorizationResponse(DataUnit query,
        String[] ids, String[] values) {
        DataUnit theQuery = (DataUnit) query;
        SAMLRequest sreq = (SAMLRequest) theQuery.getPacket();
        SAMLAuthorizationDecisionQuery q = (SAMLAuthorizationDecisionQuery) sreq.getQuery();

        try {
            Vector dataAuth = new Vector(1, 1);

            for (int i = 0; i < ids.length; i++) {
                if (ids[i].equals("AuthorizationSource")) {
                    for (int j = i + 1; j < ids.length; j++) {
                        if (ids[j].equals("AuthorizationAction")) {
                            dataAuth.addElement(values[j]);
                        } else {
                            j = ids.length;
                        }
                    }
                }
            }

            SAMLSubject rSubject = (SAMLSubject) q.getSubject();

            String[] actions = new String[dataAuth.size()];
            actions = (String[]) dataAuth.toArray(actions);

            SAMLAction samlAction = (SAMLAction) q.getActions().next();
            SAMLStatement statement;

            if (Arrays.binarySearch(actions, samlAction.getData()) >= 0) {
                statement = new SAMLAuthorizationDecisionStatement(rSubject,
                        q.getResource(), SAMLDecision.PERMIT,
                        Arrays.asList(new SAMLAction[] { samlAction }), null);
            } else {
                statement = new SAMLAuthorizationDecisionStatement(rSubject,
                        q.getResource(), SAMLDecision.DENY,
                        Arrays.asList(new SAMLAction[] { samlAction }), null);
            }

            SAMLAssertion sAssertion = new SAMLAssertion(getIssuer(),
                    new Date(), null, null, null,
                    Collections.singleton(statement));

            SAMLResponse res = null;

            res = new SAMLResponse(sreq.getId(), // ID of the Query
                    null, // URL Recipient
                    Collections.singleton(sAssertion), // Assertion
                    null // Exception
                );

            return res;
        } catch (Exception e1) {
            e1.printStackTrace();
            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                "Error creating a SAMLAuthorizationResponse");

            return null;
        }
    }

    /** Returns a <code>SAMLResponse(String inResponseTo, String recipient, Collection assertions, SAMLException e)</code> in response
     * to an authentication query, by taking into account the provided data.
     *
     * @param query A {@link DataUnit DataUnit} object
     * @param ids An array of ids needed for creating a response
     * @param values An array of ids needed linked with the ids, and also needed for creating the responsed
     *
     * @return a <code>SAMLResponse</code> object
     */
    private SAMLResponse createAuthenticationResponse(DataUnit query,
        String[] ids, String[] values) {
        DataUnit theQuery = (DataUnit) query;
        SAMLRequest sreq = (SAMLRequest) theQuery.getPacket();
        SAMLAuthenticationQuery q = (SAMLAuthenticationQuery) sreq.getQuery();

        try {
            String authMethod = null;
            String authTime = null;
            String authHost = null;

            for (int i = 0; i < ids.length; i++) {
                if (ids[i].equals("AuthenticationMethod")) {
                    authMethod = (String) values[i];
                } else if (ids[i].equals("AuthenticationTimestamp")) {
                    authTime = (String) values[i];
                } else if (ids[i].equals("AuthenticationHost")) {
                    authHost = (String) values[i];
                }
            }

            SAMLSubject rSubject = (SAMLSubject) q.getSubject();
            SAMLAuthenticationStatement as;
            SAMLAssertion sAssertion;

            if ((authTime != null) && (authHost != null)) {
                SAMLStatement statement = new SAMLAuthenticationStatement(rSubject,
                        new Date(Long.valueOf(authTime).longValue()), authHost,
                        authHost, null);

                if (authMethod != null) {
                    ((SAMLAuthenticationStatement) statement).setAuthMethod(authMethod);
                }

                sAssertion = new SAMLAssertion(getIssuer(), new Date(), null,
                        null, null, Collections.singleton(statement));
            } else {
                sAssertion = new SAMLAssertion(getIssuer(), new Date(), null,
                        null, null, null);
            }

            SAMLResponse res = null;

            res = new SAMLResponse(sreq.getId(), // ID of the Query
                    null, // URL Recipient
                    Collections.singleton(sAssertion), // Assertion
                    null // Exception
                );

            return res;
        } catch (Exception e1) {
            e1.printStackTrace();
            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                "Error creating a SAMLAuthenticationResponse");

            return null;
        }
    }

    /** Returns a <code>SAMLResponse(String inResponseTo, String recipient, Collection assertions, SAMLException e)</code> in response
     * to a query asking for attributes, by taking into account the provided data.
     *
     * @param query A {@link DataUnit DataUnit} object
     * @param ids An array of ids needed for creating a response
     * @param values An array of ids needed linked with the ids, and also needed for creating the response
     *
     * @return a <code>SAMLResponse</code> object
     */
    private SAMLResponse createAttributeResponse(DataUnit query, String[] ids,
        String[] values) {
        DataUnit theQuery = (DataUnit) query;
        SAMLRequest sreq = (SAMLRequest) theQuery.getPacket();
        SAMLAttributeQuery q = (SAMLAttributeQuery) sreq.getQuery();

        try {
            Vector dataAttr = new Vector(1, 1);

            for (int i = 0; i < ids.length; i++) {
                if (ids[i].equals("AttributeName") && (i < (ids.length - 1)) &&
                        ids[i + 1].equals("AttributeValue")) {
                    dataAttr.addElement(new String[] {
                            (String) values[i], (String) values[i + 1]
                        });
                }
            }

            SAMLAttribute[] attrs = new SAMLAttribute[dataAttr.size()];

            for (int i = 0; i < dataAttr.size(); i++) {
                String[] valueAttr = (String[]) dataAttr.elementAt(i);

                attrs[i] = new SAMLAttribute(valueAttr[0],
                        getDomainAttributeQuery(q), null, 0,
                        Arrays.asList(new String[] { valueAttr[1] }));
            }

            SAMLResponse res = null;

            SAMLSubject rSubject = (SAMLSubject) q.getSubject();
            SAMLStatement statement = new SAMLAttributeStatement(rSubject,
                    Arrays.asList(attrs));

            SAMLAssertion sAssertion = new SAMLAssertion(getIssuer(),
                    new Date(), null, null, null,
                    Collections.singleton(statement));

            res = new SAMLResponse(sreq.getId(), // ID of the Query
                    null, // URL Recipient
                    Collections.singleton(sAssertion), // Assertion
                    null // Exception
                );

            return res;
        } catch (Exception e1) {
            e1.printStackTrace();
            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                "Error creating a SAMLAttributeResponse");

            return null;
        }
    }

    /** Returns the namequalifier for a given query
     * @param query A SAML Attribute Query object
     * @return A string with the name qualifier for the query */
    private String getDomainAttributeQuery(SAMLAttributeQuery query) {
        return query.getSubject().getName().getNameQualifier();
    }

    /** Responds through the given client socket a SOAP message containing the SAML
     * response.
     *
     * @param clientSocket A client socket
     * @param response A <code>SAMLResponse</code> object that will be used to create the body of a SOAP message.
     */
    private void respond(Socket clientSocket, SAMLResponse response)
        throws java.io.IOException {
        try {
            Document doc = response.toDOM().getOwnerDocument();

            // Build the SOAP envelope and body for the response.
            Element env = doc.createElementNS(XML.SOAP11ENV_NS, "soap:Envelope");
            env.setAttributeNS(XML.XMLNS_NS, "xmlns:soap", XML.SOAP11ENV_NS);
            env.setAttributeNS(XML.XMLNS_NS, "xmlns:xsd", XML.XSD_NS);
            env.setAttributeNS(XML.XMLNS_NS, "xmlns:xsi", XML.XSI_NS);

            if (doc.getDocumentElement() == null) {
                doc.appendChild(env);
            } else {
                doc.replaceChild(env, doc.getDocumentElement());
            }

            Element body = doc.createElementNS(XML.SOAP11ENV_NS, "soap:Body");
            env.appendChild(body);

            // Attach the SAML response.
            body.appendChild(response.toDOM());

            Canonicalizer c = Canonicalizer.getInstance(Canonicalizer.ALGO_ID_C14N_OMIT_COMMENTS);

            //                  resp.setContentType("text/xml; charset=UTF-8");
            byte[] bodyMessage = c.canonicalizeSubtree(env);
            String message = getHeaderResponse(bodyMessage.length);
            clientSocket.getOutputStream().write(message.getBytes());
            clientSocket.getOutputStream().write(bodyMessage);
            clientSocket.getOutputStream().write(new String("\n").getBytes());
            clientSocket.getOutputStream().flush();
            clientSocket.close();
        } catch (InvalidCanonicalizerException ex) {
            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                "Unable to serialize XML document instance in response action");
        } catch (CanonicalizationException ex) {
            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                "Unable to serialize XML document instance in response action");
        } catch (Exception ex) {
            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                "In response action: " + ex.getClass().getName() + " " +
                ex.getMessage());
        }
    }

    /** Sends an element with the corresponding components and values
     *
     * @param name tha name of the element to be sent
     * @param components An array of strings containing the components of the element to be sent
     * @param values The values associated to the components */
    public void sendElement(String name, String[] components, String[] values)
        throws ProtocolAdaptorSendException {
        try {
            DataUnit theQuery = (DataUnit) inputQueue.getElementAt(0);

            SAMLResponse response = createResponse(theQuery, components, values);

            respond((Socket) theQuery.getPeer(), response);

            inputQueue.extractFirstElement();
        } catch (InvalidSAMLRequestException e) {
            Diagnostic.writeLog(Diagnostic.WARNING_MODE, "Request not aplicable");
            throw new ProtocolAdaptorSendException(
                "SAMLPostAdaptor: Request not applicable");
        } catch (IOException e) {
            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                "Error sending response");
            throw new ProtocolAdaptorSendException(
                "SAMLPostAdaptor: error sending response");
        }
    }

    /** Gets the values for a given param of the active element
     *
     * @param name the name of the param
     * @return An strings array with all the values for the param */
    public String[] getValue(String name) {
        DataUnit theQuery = (DataUnit) inputQueue.getElementAt(0);
        SAMLRequest sreq = (SAMLRequest) theQuery.getPacket();
        SAMLQuery q = sreq.getQuery();

        if ((name.equals("Attribute")) && (q instanceof SAMLAttributeQuery)) {
            SAMLAttributeQuery aq = (SAMLAttributeQuery) q;
            Vector v = new Vector(1, 1);

            for (Iterator i = aq.getDesignators(); i.hasNext();) {
                SAMLAttributeDesignator d = (SAMLAttributeDesignator) i.next();
                v.addElement(d.getName());
            }

            String[] res = new String[v.size()];

            return (String[]) v.toArray(res);
        } else if ((name.equals("AuthorizationSource")) &&
                (q instanceof SAMLAuthorizationDecisionQuery)) {
            SAMLAuthorizationDecisionQuery aq = (SAMLAuthorizationDecisionQuery) q;
            String[] aqv = { aq.getResource() };

            return aqv;
        } else if ((name.equals("AuthorizationAction")) &&
                (q instanceof SAMLAuthorizationDecisionQuery)) {
            SAMLAuthorizationDecisionQuery aq = (SAMLAuthorizationDecisionQuery) q;
            Iterator it = aq.getActions();
            Vector v = new Vector(1, 1);

            while (it.hasNext()) {
                SAMLAction act = (SAMLAction) it.next();
                v.addElement(act.getData());
            }

            String[] res = new String[v.size()];

            return (String[]) v.toArray(res);
        }

        return null;
    }

    /** Discards the active element from the queue and writes a diagnostic */
    public void discard() {
        Diagnostic.writeLog(Diagnostic.WARNING_MODE,
            "SAMLPostAdaptor: Discarding the active element...");

        if (inputQueue.size() > 0) {
            inputQueue.extractFirstElement();
        }
    }

    /** Sets an observer (the rule processor) for the protocol adaptor in use
     *
     * @param obs An <code>java.util.Observer</code> object */
    public void setObserver(Observer obs) throws ProtocolAdaptorConfigException {
        if (inputQueue == null) {
            Diagnostic.writeLog(Diagnostic.SEVERE_MODE,
                "SAMLPostAdaptor: Input queue has not been initialized");
            throw new ProtocolAdaptorConfigException(
                "Input queue has not been initialized");
        } else if (obs == null) {
            Diagnostic.writeLog(Diagnostic.SEVERE_MODE,
                "SAMLPostAdaptor: The provided observer is null");
            throw new ProtocolAdaptorConfigException(
                "The provided observer is null");
        } else {
            Diagnostic.writeLog(Diagnostic.INFO_MODE,
                "Adding new observer for the queue");
            inputQueue.addObserver(obs);
        }
    }

    /** Provides the label of the first available input element in the queue, and then
     * removes it from the queue; the element becomes the active one and overrides the previous.
     *
     * @return The name of the first available element in the queue */
    public String getNextInput() throws ProtocolAdaptorReadException {
        DataUnit theQuery = (DataUnit) inputQueue.getElementAt(0);
        SAMLRequest sreq = (SAMLRequest) theQuery.getPacket();
        SAMLQuery q = sreq.getQuery();

        if (q instanceof SAMLAttributeQuery) {
            return "SAMLAttributeQuery";
        } else if (q instanceof SAMLAuthenticationQuery) {
            return "SAMLAuthenticationQuery";
        } else if (q instanceof SAMLAuthorizationDecisionQuery) {
            return "SAMLAuthorizationDecisionQuery";
        } else {
            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                "SAMLPostAdaptor: The provided query (" +
                q.getClass().getName() + ") is not aplicable to SAML protocol");
            throw new ProtocolAdaptorReadException("The provided query (" +
                q.getClass().getName() + ") is not aplicable to SAML protocol");
        }
    }

    /** Sets the value for all params that are necessary for this specific
     * adaptor to work.
     * @param name is the name of the param being set
     * @param ids is an array of strings with the attribute ids for name
     * @param values is the array with the values paired with the ids array
     */
    public void setParam(String name, String[] ids, String[] values)
        throws ProtocolAdaptorConfigException {
        if (name.equals("issuer")) {
            issuerName = extractIssuerName(ids, values);
        }
    }

    /** Extra methods not in the interface */
    /** Extracts the name of the issuer from the vector of values
     *
     * @param ids An array of strings with the ids for the current request
     * @param values An array of strings with the values corresponding with the ids
     * @return An string with the name of the issuer or an empty string */
    private String extractIssuerName(String[] ids, String[] values) {
        for (int i = 0; i < ids.length; i++) {
            if (ids[i] == "name") {
                return values[i];
            }
        }

        return "";
    }

    /** Returns the name of the issuer
     *
     * @return An String with issuer's name */
    private String getIssuer() {
        return issuerName;
    }

    /** This method is used to build a String with the HTTP headers for
     * a message of the specified length
     *
     * @param len The length of the message expressed in bytes */
    private String getHeaderResponse(int len) {
        StringBuffer message = new StringBuffer("HTTP/1.0 200 OK\n");
        message.append("Content-Type: text/xml; charset=UTF-8\n");
        message.append("Content-Length: " + len);
        message.append("Connection: close\n\n");

        return message.toString();
    }
}
