/*
 * Class: RuleProcessor
 *
 * $Id: RuleProcessor.java,v 1.2 2005/05/26 11:26:55 macias Exp $
 */
package es.rediris.aarr.rules;

import es.rediris.aarr.Diagnostic;
import es.rediris.aarr.ProfileManager;
import es.rediris.aarr.protocol.ProtocolAdaptor;
import es.rediris.aarr.protocol.ProtocolAdaptorReadException;
import es.rediris.aarr.protocol.ProtocolAdaptorSendException;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import java.util.Observable;
import java.util.Observer;


/**
 *    This class is used for processing all rules in a ruleset, by checking that all conditions
 * are satisfied and then, performing the actions in every {@link Rule rule}.
 *
 * @author kan
 */
public class RuleProcessor implements Observer, Runnable {
    /** The last received request */
    static Object lastInput = null;
    static String requestName = null;
    static String[] requestValues = null;
    static int requestStatus;

    /** Request status values */
    public static final int RP_STARTING = 1;
    public static final int RP_WAITING = 2;
    public static final int RP_RECEIVED = 3;
    public static final int RP_WAIT_TIMEOUT = 4;
    public static final int RP_ENDED = 5;

    /** A {@link ProfileManager ProfileManager} */
    static ProfileManager profileManager;

    /** A boolean token to control if the thread is suspended while waiting for a request */
    private long timeout = Long.MAX_VALUE;

    /** A thread */
    Thread thread;
    RuleProcessor self;

    /** An internal {@link Ruleset} object for storing rules */
    protected Ruleset ruleSet;

    /** The main constructor receives a {@link Ruleset Ruleset} object previously loaded
     * @param ruleset A Ruleset already loaded */
    public RuleProcessor(Element ruleset) {
        parseRuleset(ruleset);

        if (!ruleSet.getName().equals("")) {
            Diagnostic.writeLog(Diagnostic.INFO_MODE,
                "Loaded ruleset " + ruleSet.getName() + " with " +
                ruleSet.getNumStates() + " states");
        } else {
            Diagnostic.writeLog(Diagnostic.INFO_MODE,
                "Loaded ruleset with " + ruleSet.getNumStates() + " states");
        }

        self = this;
        ruleSet.init();
    }

    /** Init the ruler processor by creating a {@link ProfileManager ProfileManager}, input and
     * output queues, and starting a new <code>Thread</code>
     * @param pm A {@link ProfileManager ProfileManager} Object */
    public void init(ProfileManager pm) {
        profileManager = pm;
        requestStatus = RP_STARTING;
        thread = new Thread(this, "RuleProcessor");
        thread.start();
    }

    /** Parses a document (or XML Element) that defines a rule set, loading all the states,
     *  rules, conditions and actions inside it
     *
     * @param ruleset An (XML) ruleset element previously loaded */
    private void parseRuleset(Element ruleset) {
        String nameRuleset = ruleset.getAttribute("name");

        if (nameRuleset == null) {
            nameRuleset = "";
        }

        ruleSet = new Ruleset(nameRuleset);

        NodeList listStates = ruleset.getElementsByTagName("state");

        for (int i = 0; i < listStates.getLength(); i++) {
            Element stateElem = (Element) listStates.item(i);
            String nameState = stateElem.getAttribute("name");

            if (nameState == null) {
                nameState = "";
            }

            State state = new State(nameState);
            NodeList listRules = stateElem.getElementsByTagName("rule");

            for (int j = 0; j < listRules.getLength(); j++) {
                Element ruleElem = (Element) listRules.item(j);
                String nameRule = ruleElem.getAttribute("name");

                if (nameRule == null) {
                    nameRule = "";
                }

                Rule rule = new Rule(nameRule);
                NodeList conditions = ruleElem.getElementsByTagName(
                        "conditions");

                if (conditions.getLength() > 0) {
                    NodeList listConditions = ((Element) conditions.item(0)).getElementsByTagName(
                            "condition");

                    for (int k = 0; k < listConditions.getLength(); k++) {
                        Element condElem = (Element) listConditions.item(k);
                        Condition c = searchCondition(condElem);

                        if (c == null) {
                            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                                "An unknown condition");
                        } else {
                            rule.addCondition(c);
                        }
                    }
                }

                NodeList actions = ruleElem.getElementsByTagName("actions");

                if (actions.getLength() > 0) {
                    NodeList listActions = ((Element) actions.item(0)).getElementsByTagName(
                            "action");

                    for (int k = 0; k < listActions.getLength(); k++) {
                        Element actionElem = (Element) listActions.item(k);
                        Action a = searchAction(actionElem);

                        if (a == null) {
                            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                                "An unknown action");
                        } else {
                            rule.addAction(a);
                        }
                    }
                }

                state.addRule(rule);
            }

            ruleSet.addState(state);
        }
    }

    /** Match the kind of condition being processed at the moment
     * and returns a condition defined in element (<code>org.w3c.dom.Element</code> Object)
     * of the document.
     * @param condElem An element (<code>org.w3c.dom.Element</code>) defining a condition
     * @return A {@link Condition Condition} object: one of {@link ConditionDefault ConditionDefault},
     * {@link ConditionField ConditionField}, or {@link ConditionReceive ConditionReceive}. */
    private Condition searchCondition(Element condElem) {
        if (condElem.hasAttribute("receive")) {
            return new ConditionReceive(condElem);
        } else if (condElem.hasAttribute("field")) {
            return new ConditionField(condElem);
        } else if (condElem.hasAttribute("default")) {
            return new ConditionDefault(condElem);
        } else {
            return null;
        }
    }

    /** Adds a new query to the input queue
     *
     * @param obs An {@link Observable ObservableVector} with queries
     * @param data @deprecated Not used */
    public synchronized void update(Observable obs, Object data) {
        Diagnostic.writeLog(Diagnostic.INFO_MODE, "New query");

        try {
            requestName = profileManager.getProtocolAdaptor().getNextInput();
            requestValues = profileManager.getProtocolAdaptor().getValue(requestName);
            lastInput = requestName;

            notify();
        } catch (ProtocolAdaptorReadException e) {
            Diagnostic.writeLog(Diagnostic.INFO_MODE,
                "Unable to determine next input element");
        }
    }

    /** Removes last query from the input queue */
    protected static void dropQuery() {
        Diagnostic.writeLog(Diagnostic.INFO_MODE, "Last query dropped");

        // discard aqu
    }

    /** Send a response back to the <code>ProtocolAdaptor</code>
     *
     * @param ids An array of ids
     * @param values The corresponding array of values for the ids */
    protected static void putOutput(String[] ids, String[] values) {
        ProtocolAdaptor man = profileManager.getProtocolAdaptor();

        try {
            man.sendElement(requestName, ids, values);
            lastInput = null;
            requestName = null;
        } catch (ProtocolAdaptorSendException e) {
            Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                "Unable to send response for request '" + requestName + "'.");

            // + discard()
        }
    }

    /** Similar to {@link RuleProcessor#searchCondition(Element ) searchCondition()} but for Actions.
     * Matches the kind of action being processed at the moment returning the action defined
     * in an element (<code>org.w3c.dom.Element</code> Object) of the ruleset being processed.
     * @param actionElem An element (<code>org.w3c.dom.Element</code>) defining an action
     * @return A {@link Action Action} object: one of {@link ActionSend ActionSend}, {@link ActionNext ActionNext},
     * or {@link ActionFinish ActionFinish } */
    private Action searchAction(Element actionElem) {
        if (actionElem.hasAttribute("next")) {
            return new ActionNext(actionElem);
        } else if (actionElem.hasAttribute("send")) {
            return new ActionSend(actionElem);
        } else if (actionElem.hasAttribute("finish")) {
            return new ActionFinish(actionElem);
        } else {
            return null;
        }
    }

    /** Returns a <code>String</code> with the name of the last input element
     * (the first in the input query) */
    protected static String getInputQuery() {
        try {
            return (String) profileManager.getProtocolAdaptor().getNextInput();
        } catch (ProtocolAdaptorReadException e) {
            Diagnostic.writeLog(Diagnostic.INFO_MODE,
                "Unable to determine next input's name");
        }

        return null;
    }

    /** Gets the value of a param (defined by <b>data</b>) for the protocol adaptor in use
     * @param name The param to ask for
     * @return The value associated with the param expressed by data */
    protected static String[] getParam(String name) {
        return profileManager.getProtocolAdaptor().getValue(name);
    }

    /** Sets the end execution token to true */
    protected static void endExecution() {
        requestStatus = RP_ENDED;
    }

    /** Checks if the rule processor has finished its execution.
     * @return True if execution has finished, false otherwise. */
    protected static boolean isFinished() {
        return (requestStatus == RP_ENDED) ? true : false;
    }

    /** Sets the timemout for the rule processor
     * @param millisecs Time in milliseconds */
    public void setTimeout(long millisecs) {
        timeout = millisecs;
    }

    /** Executes actual states and sends information about the current status
     * to the {@link Diagnostic diagnostics module} */
    public void run() {
        while (true) {
            if (requestStatus == RP_ENDED) {
                Diagnostic.writeLog(Diagnostic.INFO_MODE,
                    "We've just reached a final state");

                ruleSet.init();
                requestStatus = RP_STARTING;
                Diagnostic.writeLog(Diagnostic.INFO_MODE,
                    "Rule processor processor reset to its initial state");
            } else {
                self.waitInput(timeout);

                Diagnostic.writeLog(Diagnostic.INFO_MODE, "Processing input...");
                ruleSet.executeCurrentState();
            }
        }
    }

    /** Wait for a request the amount of time provided as a param
     *
     * @param millisecs Time in milliseconds to wait until a request is received or a timeout condition is reached */
    public synchronized void waitInput(long millisecs) {
        Diagnostic.writeLog(Diagnostic.INFO_MODE, "Waiting for a query");
        requestStatus = RP_WAITING;

        if (ruleSet.needsAQuery() && (requestName == null)) {
            try {
                if (millisecs > 0) {
                    wait(millisecs);
                } else {
                    wait();
                }
            } catch (InterruptedException e) {
                Diagnostic.writeLog(Diagnostic.INFO_MODE, "Received a request");
                requestStatus = RP_RECEIVED;
            }

            requestStatus = RP_WAIT_TIMEOUT;
        }
    }
}
