/*
 * Class: Ruleset
 *
 * $Id: Ruleset.java,v 1.1.1.1 2005/05/22 17:00:49 macias Exp $
 */
package es.rediris.aarr.rules;

import es.rediris.aarr.Diagnostic;

import java.util.Iterator;
import java.util.Vector;


/**
 * This class defines a Ruleset (ie: A set of rules) and is used to store states
 * and execute the actions for every state depending on validity of the conditions
 * for every {@link Rule Rule}.
 *
 * @author kan
 */
public class Ruleset {
    /** A vector for storing the states of the ruleset */
    Vector states;

    /** The name of the ruleset */
    String name;

    /** The actual {@link State State} */
    State actualState;

    /** The main constructor creates a new <code>Java.lang.Vector</code> to store the
     states associated with this Ruleset */
    public Ruleset() {
        states = new Vector(1, 1);
    }

    /** This other constructor gives a name to the Ruleset after creating a new
     <code>Java.lang.Vector</code> to store the states associated with this Ruleset
    @param name The name for the Ruleset */
    public Ruleset(String name) {
        this();
        setName(name);
    }

    /** Gets the name of this Ruleset
    @return The name of the ruleset */
    public String getName() {
        return name;
    }

    /** Sets the name of this Ruleset
    @param name The name for the ruleset */
    public void setName(String name) {
        this.name = name;
    }

    /** Returns an <code>java.util.Iterator</code> from the vector containing the state in the current Ruleset.
    @return A <code>java.util.Iterator</code> object */
    public Iterator getIterator() {
        return states.iterator();
    }

    /** Adds a new State to the {@link Ruleset#states states} vector
    @param state A {@link State State} object */
    public void addState(State state) {
        states.addElement(state);
    }

    /** Removes a state at the given position from the {@link Ruleset#states states} vector
    @param index Position of the states to be removed */
    public void delState(int index) {
        states.remove(index);
    }

    /** Returns the number of states in this Ruleset
    @return An integer */
    public int getNumStates() {
        return states.size();
    }

    /** Gets the state at the given position of the {@link Ruleset#states states} vector
    @param index Position of the state
    @return A {@link State State} object */
    public State getState(int index) {
        return (State) states.elementAt(index);
    }

    /** Gets the {@link State State} by its name from the {@link Ruleset#states states} vector
    @param name The name of the state
    @return A {@link State State} object */
    public State getState(String name) {
        for (int i = 0; i < getNumStates(); i++) {
            State state = (State) states.elementAt(i);

            if (state.getName().equals(name)) {
                return state;
            }
        }

        return null;
    }

    /** Changes to the first {@link State State} or displays an error if there are no states */
    public void init() {
        if (getNumStates() > 0) {
            changeState(0);
        } else {
            Diagnostic.writeLog(Diagnostic.WARNING_MODE, "There are no states!");
        }
    }

    /** Changes to the {@link State State} specified by the index param and displays
       the action through the {@link Diagnostic diagnostics module}.
    @param index The position of the state */
    protected void changeState(int index) {
        actualState = getState(index);

        String nameState = actualState.getName();

        if (nameState.equals("")) {
            Diagnostic.writeLog(Diagnostic.INFO_MODE,
                "Changing to state #" + index);
        } else {
            Diagnostic.writeLog(Diagnostic.INFO_MODE,
                "Changing to '" + nameState + "' state");
        }
    }

    /** Changes to the {@link State State} specified by the name param and displays
       the action through the {@link Diagnostic diagnostics module}.
    @param nameState The name of the {@link State State} we are changing to */
    protected void changeState(String nameState) {
        actualState = getState(nameState);
        RuleProcessor.dropQuery();
        Diagnostic.writeLog(Diagnostic.INFO_MODE,
            "Changing to '" + nameState + "' state");
    }

    /** Checks if next condition inside the actual {@link State State} need a query or not.
    @return True if the condition needs a query, false otherwise */
    public boolean needsAQuery() {
        boolean res = false;

        Iterator iter = actualState.getIterator();

        while (iter.hasNext()) {
            Rule rule = (Rule) iter.next();
            Iterator iterCond = rule.getConditions();

            while (iterCond.hasNext()) {
                Condition cond = (Condition) iterCond.next();
                res |= cond.needsAQuery();
            }
        }

        return res;
    }

    /** Executes actual {@link State State}, checking also if this is a final one. */
    public void executeCurrentState() {
        String nameNewState = actualState.executeState();

        if (!RuleProcessor.isFinished()) {
            if ((nameNewState != null) && (getState(nameNewState) == null)) {
                Diagnostic.writeLog(Diagnostic.WARNING_MODE,
                    "'" + nameNewState + "' state not found");
                RuleProcessor.endExecution();
            } else if ((nameNewState != null) &&
                    !actualState.getName().equals(nameNewState)) {
                changeState(nameNewState);
            } else {
                Diagnostic.writeLog(Diagnostic.INFO_MODE,
                    "Again '" + actualState.getName() + "' state");
            }
        }
    }
}
